hidden_alphabet = "\u200B\u200C\u200D\u202F"
hidden_alphabet_table = {
    "\u200B": 0,
    "\u200C": 1,
    "\u200D": 2,
    "\u202F": 3,
}


def encode_byte(b):
    r = ""
    for _ in range(4):
        r = r + hidden_alphabet[b % 4]
        b //= 4
    return r


def decode_hidden_byte(h: str):
    b = 0
    for i, hb in enumerate(h):
        b |= hidden_alphabet_table[hb] << i * 2
    return b


def encode(ss: str):
    r = ""
    for b in ss.encode("ascii"):
        r = r + encode_byte(b)
    return r


def decode(ss: str):
    r = bytearray()
    for b in range(len(ss) // 4):
        r.append(decode_hidden_byte(ss[b * 4 : b * 4 + 4]))
    return r.decode("ascii")



from __future__ import absolute_import

import itertools
from typing import Union


def chunks(l: list, n: int):
    for i in range(0, len(l), n):
        yield l[i : i + n]


def ichunks(iterable: Union[list, tuple], n: int):
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk



import random
import string


def random_string(length):
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def generate_request_id(length: int = 12):
    return random_string(length)


async def reservoir_sampling_async(async_iterator, n):
    pool = []
    index = 0
    async for el in async_iterator:
        if index < n:
            pool.append(el)
        else:
            r = random.randint(0, index)
            if r < n:
                pool[r] = el
        index += 1
    return pool



def import_object(name):
    components = name.split(".")
    mod = __import__(components[0])
    parts = [components[0]]
    for comp in components[1:]:
        parts.append(comp)
        try:
            mod = __import__(".".join(parts), fromlist=[""])
        except (ImportError, ModuleNotFoundError):
            mod = getattr(mod, comp)
    return mod


def instantiate_object(descriptor):
    return import_object(descriptor["class"])(**descriptor.get("kwargs", {}))


def class_fullname(o):
    klass = o.__class__
    module = klass.__module__
    if module == "builtins":
        return klass.__qualname__
    return module + "." + klass.__qualname__



import logging

from .tskv import TskvRecord


class BaseError(Exception):
    code = "base_error"
    level = logging.ERROR
    info = dict()
    internal_info = dict()

    def __init__(self, **kwargs):
        self.info = dict(self.info)
        self.internal_info = dict(self.internal_info)

        if "code" in kwargs:
            self.code = kwargs["code"]

        if "internal_info" in kwargs:
            self.internal_info.update(kwargs.pop("internal_info"))

        self.info["code"] = self.code
        self.info.update(kwargs)

    def as_dict(self):
        return dict(self.info)

    def as_tskv(self):
        return TskvRecord(self.info)

    def as_internal_dict(self):
        info = self.as_dict()
        info.update(self.internal_info)
        return info

    def as_internal_tskv(self):
        info = self.as_tskv()
        info.update(self.internal_info)
        return info

    def add(self, key, value):
        self.info[key] = value

    def __str__(self):
        return f"{self.__class__.__name__}({self.as_dict()})"


class MissingFileError(BaseError, FileNotFoundError):
    def __init__(self, file_path=""):
        self.info = {"path": file_path}


class NeedRetryError(BaseError):
    pass



import hashlib

b58 = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def base58_encode(n):
    result = ""
    while n > 0:
        result = b58[n % 58] + result
        n //= 58
    return result


def base58_decode(s):
    result = 0
    for i in range(0, len(s)):
        result = result * 58 + b58.index(s[i])
    return result


def base256_encode(n):
    result = b""
    while n > 0:
        result = bytes([n % 256]) + result
        n //= 256
    return result


def base256_decode(s):
    result = 0
    for c in s:
        result = result * 256 + c
    return result


def count_leading_chars(s, ch):
    count = 0
    for c in s:
        if c == ch:
            count += 1
        else:
            break
    return count


# https://en.bitcoin.it/wiki/Base58Check_encoding
def base58_check_encode(payload, version_byte=0):
    s = bytes([version_byte]) + payload
    checksum = hashlib.sha256(hashlib.sha256(s).digest()).digest()[0:4]
    result = s + checksum
    leading_zeros = count_leading_chars(result, 0)
    return "1" * leading_zeros + base58_encode(base256_decode(result))


def base58_check_decode(s):
    leading_ones = count_leading_chars(s, "1")
    s = base256_encode(base58_decode(s))
    result = b"\0" * leading_ones + s[:-4]
    chk = s[-4:]
    checksum = hashlib.sha256(hashlib.sha256(result).digest()).digest()[0:4]
    assert chk == checksum
    return result[1:], result[0]


def reverse_hex(s):
    return bytes.fromhex(s)[::-1].hex()



import enum
import types
from datetime import date, datetime

from .common import filter_none


class JsonSerializator:
    def __init__(
        self,
        max_serialization_depth,
        custom_field_processors=None,
        custom_type_processors=None,
    ):
        self._max_serialization_depth = max_serialization_depth
        self._custom_field_processors = custom_field_processors or {}
        self._custom_type_processors = custom_type_processors or {}

    def _process_list(
        self, field, depth, excludes=None, includes=None, *args, **kwargs
    ):
        depth += 1
        if depth > self._max_serialization_depth:
            return

        r = [
            self.process_field(
                x, depth, excludes=excludes, includes=includes, *args, **kwargs
            )
            for x in field
        ]
        return filter_none(r)

    def _process_dict(
        self, field, depth, excludes=None, includes=None, *args, **kwargs
    ):
        depth += 1
        result = {}
        if depth > self._max_serialization_depth:
            return result
        for key in field:
            if excludes and key in excludes:
                continue
            if key in self._custom_field_processors:
                value = self._custom_field_processors[key](
                    field[key],
                    depth,
                    excludes=excludes,
                    includes=includes,
                    *args,
                    **kwargs
                )
            else:
                value = self.process_field(
                    field[key],
                    depth,
                    excludes=excludes,
                    includes=includes,
                    *args,
                    **kwargs
                )
            if value is None:
                continue
            key = self.process_field(
                key, depth, excludes=excludes, includes=includes, *args, **kwargs
            )
            result[key] = value
        return result

    def process_field(
        self, field, depth, excludes=None, includes=None, *args, **kwargs
    ):
        for type_, processor in self._custom_type_processors.items():
            if isinstance(field, type_):
                return processor(
                    field, depth, excludes=excludes, includes=includes, *args, **kwargs
                )
        if isinstance(field, (date, datetime)):
            return field.strftime("%Y-%m-%d %H:%M:%S")
        if isinstance(field, (enum.Enum, enum.IntEnum)):
            return field.name
        if isinstance(field, (list, tuple, set)) or isinstance(
            field, types.GeneratorType
        ):
            return self._process_list(
                field, depth, excludes=excludes, includes=includes, *args, **kwargs
            )
        if isinstance(field, dict):
            return self._process_dict(
                field, depth, excludes=excludes, includes=includes, *args, **kwargs
            )
        return field


class JsonSerializable:
    serializator = JsonSerializator
    max_serialization_depth = 1

    def serialize(self, max_depth=None, excludes=None, includes=None, *args, **kwargs):
        serializator = self.serializator(
            max_serialization_depth=max_depth or self.max_serialization_depth
        )
        return serializator.process_field(
            self, depth=0, excludes=excludes, includes=includes, *args, **kwargs
        )



import os
import platform
import socket

hostname = socket.gethostname()
node_name = os.environ.get("NODE_NAME")
type = os.environ.get("ENV_TYPE", "development")
name = os.environ.get("ENV_TYPE", "development")
platform = platform.system().lower()



MAX_LOG_LENGTH = 8192
REPLACE_MAP = [
    ("\\", r"\\"),
    ("\t", r"\t"),
    ("\n", r"\n"),
    ("\r", r"\r"),
    ("\0", r"\0"),
]


class TskvRecord(dict):
    preferred_fields = ["unixtime", "datetime", "request_id", "client_ip", "path"]

    @staticmethod
    def escape_value(text):
        for char, replacement in REPLACE_MAP:
            text = text.replace(char, replacement)
        return text

    @staticmethod
    def truncate_str(text):
        return text[:MAX_LOG_LENGTH]

    def __str__(self):
        elements = []
        for preferred_field in self.preferred_fields:
            if self.get(preferred_field) is not None:
                elements.append(
                    self.escape_value(
                        "%s=%s" % (preferred_field, self[preferred_field])
                    )
                )
                self.pop(preferred_field)
        for k, v in sorted(self.items()):
            if v is not None:
                elements.append(self.escape_value("%s=%s" % (k, v)))
        return self.truncate_str("\t".join(elements))
